import multiprocessing
import threading
import numpy as np
import talib
from ks_trade_api.constant import Adjustment, RET_OK, SUBSCRIBE_TYPE2INTERVAL, IndicatorColumn, Interval, Indicator, Timing, SubscribeType
from ks_trade_api.object import BarData
from ks_futu_market_api import KsFutuMarketApi
from ks_utility.zmqs import ZmqPublisher, ZmqSubscriber
from logger import Logger
import time
from pathlib import Path
import json
import os
from setting import SETTINGS
from ks_trade_api.utility import TEMP_DIR

PARAMS2ENUM: dict = {
    'types': SubscribeType,
    'indicators': Indicator
}

def process_b(bar: BarData, close_shared_dict: dict, datetime_shared_dict: dict, indicator_shared_queue: multiprocessing.Queue):
    """ 计算不同周期的 price 均值 """
    vt_symbol = bar.vt_symbol
    interval = bar.interval  # 取出当前数据的周期
    
    history_updated = False
    history_datetime = None
    if datetime_shared_dict[vt_symbol][interval][-1] < bar.datetime:
        history_datetime = datetime_shared_dict[vt_symbol][interval][-1]
        print(f'[{vt_symbol},{interval}]history_updated!!!!!!!!!!!!!!@{history_datetime}')
        close_shared_dict[vt_symbol][interval]._callmethod("append", (float(bar.close),))
        datetime_shared_dict[vt_symbol][interval]._callmethod("append", (bar.datetime,))
        close_shared_dict[vt_symbol][interval][:] = close_shared_dict[vt_symbol][interval][-21:]
        datetime_shared_dict[vt_symbol][interval][:] = datetime_shared_dict[vt_symbol][interval][-21:]
        history_updated = True
    else:
        close_shared_dict[-1] = float(bar.close)

    close_prices = np.array(close_shared_dict[vt_symbol][interval])
    bars = []
    boll_band = talib.BBANDS(close_prices[-20:], timeperiod=20)

    if history_updated:
        # todo! 为了性能考虑，先不要广播realtime_bar了
        # realtime_bar = {
        #     'symbol': bar.symbol, 
        #     'exchange': bar.exchange.value,
        #     'datetime': str(bar.datetime),
        #     'interval': interval.value, 
        #     'indicator': Indicator.BAR.value,
        #     'timing': Timing.REALTIME.value, 
        #     'data': {
        #         IndicatorColumn.BARCLOSE.name: close_prices[-1]
        #     }
        # }
        
        history_bar = {
            'symbol': bar.symbol, 
            'exchange': bar.exchange.value,
            'datetime': str(history_datetime),
            'interval': interval.value, 
            'indicator': Indicator.BAR.value,
            'timing': Timing.HISTORY.value, 
            'data': {
                IndicatorColumn.BARCLOSE.name: close_prices[-2]
            }
        }
        bars.append(history_bar)
        
        # pdb.set_trace()
        boll_band = talib.BBANDS(close_prices, timeperiod=20)
        realtime_boll = {
            'symbol': bar.symbol, 
            'exchange': bar.exchange.value,
            'datetime': str(bar.datetime),
            'interval': interval.value, 
            'indicator': Indicator.BOLL.value,
            'timing': Timing.REALTIME.value, 
            'data': {
                IndicatorColumn.BOLLUPPER.name: boll_band[0][-1], 
                IndicatorColumn.BOLLMIDDLE.name: boll_band[1][-1], 
                IndicatorColumn.BOLLLOWER.name: boll_band[2][-1]
            }
        }
        history_boll = {
            'symbol': bar.symbol, 
            'exchange': bar.exchange.value,
            'datetime': str(history_datetime),
            'interval': interval.value, 
            'indicator': Indicator.BOLL.value,
            'timing': Timing.HISTORY.value, 
            'data': {
                IndicatorColumn.BOLLUPPER.name: boll_band[0][-2],
                IndicatorColumn.BOLLMIDDLE.name: boll_band[1][-2], 
                IndicatorColumn.BOLLLOWER.name: boll_band[2][-2]
            }
        }
        bars.append(realtime_boll)
        bars.append(history_boll)
    else:
        # pdb.set_trace()
        boll_band = talib.BBANDS(close_prices[-20:], timeperiod=20)
        realtime_boll = {
            'symbol': bar.symbol, 
            'exchange': bar.exchange.value,
            'datetime': str(bar.datetime),
            'interval': interval.value, 
            'indicator': Indicator.BOLL.value,
            'timing': Timing.REALTIME.value, 
            'data': {
                IndicatorColumn.BOLLUPPER.name: boll_band[0][-1],
                IndicatorColumn.BOLLMIDDLE.name: boll_band[1][-1],
                IndicatorColumn.BOLLLOWER.name: boll_band[2][-1]
            }
        }
        bars.append(realtime_boll)
    
    if bars:
        indicator_shared_queue.put(bars)

    # pid = os.getpid()  # 获取当前进程 ID
    # print(f"[进程 {pid}] 股票: {vt_symbol} | 周期: {interval}")


class KsMarketApiServer(ZmqPublisher, ZmqSubscriber, Logger):
    gateway_name: str = 'ks_market_api'
    
    def __init__(self, setting: dict):
        ZmqPublisher.__init__(self, setting['zmq']['pub_address'], server=True)
        ZmqSubscriber.__init__(self, setting['zmq']['sub_address'], server=True)
        
        log_path = TEMP_DIR.joinpath('logs').joinpath('ks_trader_api')
        Logger.__init__(self, path=log_path, name='ks_market_api')
        
        self.log('KsMarketApiServer初始化...')

        self.gateway_name = setting.get('gateway_name', self.gateway_name)
        self.market_api = KsFutuMarketApi(setting)

        self.running = True
        self.pool = multiprocessing.Pool(processes=setting.get('pool_size', 1))
        self.manager = multiprocessing.Manager()
        self.close_shared_dict = self.manager.dict() # 储存股票数据 股票->周期->收盘价
        self.datetime_shared_dict = self.manager.dict() # 储存股票数据 股票->周期->时间
        self.indicator_shared_queue = self.manager.Queue() # 储存计算好的布林带值
        self.bar_shared_queue = self.manager.Queue() # 储存计算好的布林带值

        def on_bar_rt(bar: BarData):
            self.bar_shared_queue.put(bar)

        self.market_api.on_bar_rt = on_bar_rt
        
        # 启动后台线程运行主循环
        self.indicator_thread = threading.Thread(target=self.process_indicator)
        self.indicator_thread.start()
        
        self.log('KsMarketApiServer初始化完成.')
        
        self.process_bar()
        
        # todo!!! remove
        # from ks_trade_api.constant import Exchange, Interval, SubscribeType
        # from ks_trade_api.utility import datetimes
        # from ks_utility.datetimes import get_dt_int
        # self.subscribe(['00700.SEHK'], [SubscribeType.K_MINUTE])
        # while True:
        #     self.market_api.on_bar_rt(BarData(
        #         gateway_name=self.gateway_name,
        #         symbol='00700',
        #         exchange=Exchange.SEHK,
        #         datetime=get_dt_int(interval=60),
        #         interval=Interval.MINUTE,
        #         volume=100,
        #         turnover=100,
        #         open=199,
        #         high=2003,
        #         low=1023,
        #         close=533.5
        #     ))
        #     import time
        #     time.sleep(10)
        
    def update_shared_dict(self, vt_symbol: str, interval: Interval):
        if not vt_symbol in self.close_shared_dict:
            self.close_shared_dict[vt_symbol] = self.manager.dict()
        if not interval in self.close_shared_dict[vt_symbol]:
            self.close_shared_dict[vt_symbol][interval] = self.manager.list()
            
        if not vt_symbol in self.datetime_shared_dict:
            self.datetime_shared_dict[vt_symbol] = self.manager.dict()
        if not interval in self.datetime_shared_dict[vt_symbol]:
            self.datetime_shared_dict[vt_symbol][interval] = self.manager.list()

    def on_message(self, topic, msg):
        msg_data = json.loads(msg)
        # todo 以后可能不是都传入list
        for param, value_list in msg_data.items():
            if param in PARAMS2ENUM:
                msg_data[param] = [PARAMS2ENUM[param](x) for x in value_list]
        if topic == 'close':
            self.close()
        elif topic == 'subscribe':
            self.subscribe(**msg_data)

    def subscribe(
            self,
            vt_symbols: list[str] = [], 
            types: list[SubscribeType] = [], 
            indicators: list[Indicator] = [],
            data_time_types: list[Timing] = []
        ):
        api = self.market_api
        # 订阅市场数据
        api.subscribe(vt_symbols, types)
        for subscribe_type in types:
            for vt_symbol in vt_symbols:
                interval = SUBSCRIBE_TYPE2INTERVAL.get(subscribe_type)
                if not interval == Interval.TICK:
                    ret_k, data_k = api.query_history_n(vt_symbol, 20, interval=interval, adjustment=Adjustment.BACKWARD_ADJUSTMENT)
                    if ret_k == RET_OK:
                        self.log(f'{vt_symbol},{interval.value}历史bars查询成功')
                        self.update_shared_dict(vt_symbol, interval)
                        self.close_shared_dict[vt_symbol][interval]._callmethod('extend', (data_k.close.tolist(),))
                        self.datetime_shared_dict[vt_symbol][interval]._callmethod('extend', (data_k.datetime.tolist(),))

    def process_bar(self):
        """ 处理布林带计算结果 """
        
        while self.running:
            bar = self.bar_shared_queue.get()
            if bar is None:
                break
            
            self.pool.apply_async(process_b, (bar, self.close_shared_dict, self.datetime_shared_dict, self.indicator_shared_queue))  # 提交任务到进程池

    def process_indicator(self):
        """ 处理布林带计算结果 """
        while self.running:
            bars = self.indicator_shared_queue.get()
            if bars is None:
                break
            
            for bar in bars:
                bar['gateway_name'] = self.gateway_name
                print(bar) # todo!!!
                self.send('on_indicator', bar)

    def close(self):
        """ 关闭所有进程 """
        
        self.bar_shared_queue.put(None)
        self.indicator_shared_queue.put(None)
        self.indicator_thread.join()
        
        self.running = False
        self.pool.close()
        self.pool.join()
        
        self.send('on_close', {})
        
        ZmqPublisher.stop(self)
        ZmqSubscriber.stop(self)
        
        self.log('全部线程进程已经关闭.')


if __name__ == '__main__':
    multiprocessing.freeze_support() # 支持pyinstaller打包多进程

    api_setting = SETTINGS['ks_market_api']['setting']
    KsMarketApiServer(api_setting)



from threading import RLock, Thread
from datetime import datetime
from time import sleep
from futu import OpenQuoteContext
from futu.common.open_context_base import OpenContextBase, ContextStatus
from futu.common.callback_executor import CallbackExecutor, CallbackItem
from futu.common.network_manager import NetManager, PacketErr, ConnectErr, CloseReason
from futu.common.handler_context import HandlerContext
from futu.common.sys_config import SysConfig
from futu.common.err import Err
from futu.common.constant import RET_OK
from futu.quote.open_quote_context import SubRecord

class KsOpenContextBase(OpenContextBase):
    def __init__(self, host, port, is_encrypt=None, is_async_connect=False):
        self.__host = host
        self.__port = port
        self._callback_executor = CallbackExecutor()
        self._net_mgr = NetManager.default()
        self._handler_ctx = HandlerContext(self._is_proc_run)
        self._lock = RLock()
        self._status = ContextStatus.START
        self._connect_err = None  # rsa加密失败时为Err.RsaErr, 否则为str
        self._proc_run = True
        self._sync_conn_id = 0
        self._conn_id = 0
        self._keep_alive_interval = 10
        self._last_keep_alive_time = datetime.now()
        self._reconnect_timer = None
        self._reconnect_interval = 8  # 重试连接的间隔
        self._sync_query_connect_timeout = None
        self._keep_alive_fail_count = 0
        self._last_recv_time = datetime.now()
        self._is_encrypt = is_encrypt
        if self.is_encrypt():
            assert SysConfig.INIT_RSA_FILE != '', Err.NotSetRSAFile.text
        self._net_mgr.start()

        if is_async_connect:
            self._wait_reconnect(0)
        else:
            while True:
                ret = self._init_connect_sync()
                if ret == RET_OK:
                    return
                else:
                    if self.status == ContextStatus.CLOSED:
                        return
                sleep(self._reconnect_interval)
    
class KsOpenQuoteContext(KsOpenContextBase):
    def __init__(self, host='127.0.0.1', port=11111, is_encrypt=None, is_async_connect=False):
        """
        初始化Context对象
        :param host: host地址
        :param port: 端口
        """
        self._ctx_subscribe = {}
        self._sub_record = SubRecord()
        super(KsOpenContextBase, self).__init__(
            host, port, is_encrypt=is_encrypt, is_async_connect=is_async_connect)


from ks_utility.logs import LoggerBase
from pathlib import Path

class Logger(LoggerBase):
    def __init__(self, path=None, name=None):
        if not path:
            path = Path('data/logs')
        if not name:
            name = 'ks_market_engine'
        super().__init__(path, name)


# SPDX-FileCopyrightText: 2024-present liuyanzhi08 <702368372@qq.com>
#
# SPDX-License-Identifier: MIT
# __version__ = "0.1.66" # 周期增加月份支持
# __version__ = "1.0.0" # 移除对老虎api依赖
__version__ = "1.0.3" # query_hisotry_n由decimal改为返回float
__version__ = "1.0.4" # get_trading_time默认返回原来交易所时区


from sympy import sympify, lambdify
from sympy.logic.boolalg import Boolean
from pandas import DataFrame
from typing import Union
import numpy as np
from .object import BaseException

class MathsException(BaseException):
    pass

class ExpressionCache:
    """用于缓存表达式和对应矢量化函数的类"""
    _cache = {}

    @classmethod
    def get_calculator(cls, expression):
        """获取或创建表达式的矢量化计算器"""
        if expression not in cls._cache:
            cls._cache[expression] = VectorizedExpressionCalculator(expression)
        return cls._cache[expression]


class VectorizedExpressionCalculator:
    """矢量化表达式计算类"""
    def __init__(self, expression):
        # 将表达式字符串转换为 SymPy 表达式
        self.expression = sympify(expression)
        # 将 SymPy 表达式转换为矢量化函数（支持 NumPy/Pandas 运算）
        self.variables = sorted(self.expression.free_symbols, key=lambda x: str(x)) # 这里要用sorted很关键，因为set是没有顺序的，会导致值计算错误
        self.vectorized_func = lambdify(
            self.variables,
            self.expression,
            modules=[{
                'max': np.maximum,
                'min': np.minimum,
                '&': np.logical_and,
                '|': np.logical_or,
                '==': np.equal,
                '>': np.greater,
                '<': np.less,
                '>=': np.greater_equal,
                '<=': np.less_equal
            }, 'numpy']
        )

    def calculate_dataframe(self, df):
        # 确保 DataFrame 中包含表达式需要的所有变量
        required_columns = [str(x) for x in self.variables]
        missing_columns = set(required_columns) - set(df.columns)
        if missing_columns:
            raise ValueError(f"Missing required columns: {missing_columns}")
        
        # 使用矢量化函数计算
        return self.vectorized_func(*(df[col] for col in required_columns))


def calculate_expression(expression, df):
    """接口函数，直接传入表达式和 DataFrame 进行计算"""
    calculator = ExpressionCache.get_calculator(expression)
    return calculator.calculate_dataframe(df)

def cal_exp(exp: str, values: Union[dict, DataFrame]):
    try:
        if isinstance(values, dict):
            # 将表达式字符串转换为 SymPy 表达式
            expression = sympify(exp)
            expression = expression.subs(values)
            # 如果结果是布尔值（BooleanTrue 或 BooleanFalse）
            if isinstance(expression, Boolean):
                return bool(expression)  # 返回Python的布尔类型
            # 如果结果是浮动类型，计算并返回浮动值
            else:
                return float(expression.evalf()) if expression.is_number else expression
        else:
            calculator = ExpressionCache.get_calculator(exp)
            return calculator.calculate_dataframe(values)
    except Exception as e:
        raise MathsException(f"表达式{exp}计算错误：{str(e)}", data={'values': values})

# t = cal_exp('a+b+c', {'a': 1, 'b': 2, 'c': 3})
# print(t)


from futu import *
from time import sleep
from datetime import datetime, timedelta
import numpy as np
import pandas as pd
from pandas import DataFrame
from ks_trade_api.utility import extract_vt_symbol, generate_vt_symbol, is_stock_like, get_trading_time
from ks_trade_api.constant import (
    CHINA_TZ, US_EASTERN_TZ,
    Exchange as KsExchange, Product as KsProduct, SubscribeType as KsSubscribeType,
    RetCode as KsRetCode, RET_OK as KS_RET_OK, RET_ERROR as KS_RET_ERROR, ErrorCode as KsErrorCode,
    Interval as KsInterval, Adjustment as KsAdjustment
)
from ks_trade_api.object import (
    ErrorData, ContractData, MyTickData, MyBookData, MyRawTickData, QuoteData, BarData
)
from ks_utility import datetimes
from decimal import Decimal
from dateutil.parser import parse
from ks_trade_api.base_market_api import BaseMarketApi
from typing import Optional, Union, List
from logging import DEBUG, INFO, WARNING, ERROR
from futu.common.open_context_base import ContextStatus

PRODUCT_MY2KS = {
    SecurityType.STOCK: KsProduct.EQUITY,
    SecurityType.DRVT: KsProduct.OPTION
}

RET_KS2MY = {
    KS_RET_OK: RET_OK,
    KS_RET_ERROR: RET_ERROR
}

RET_MY2KS = { v:k for k,v in RET_KS2MY.items() }

MARKET_KS2MY = {
    KsExchange.SEHK: TrdMarket.HK,
    KsExchange.SMART: TrdMarket.US,
    KsExchange.SSE: 'SH',
    KsExchange.SZSE: 'SZ'
}

MARKET_MY2KS = { v:k for k,v in MARKET_KS2MY.items() }

PRODUCT_KS2MY = {
    KsProduct.EQUITY: SecurityType.STOCK,
    KsProduct.OPTION: SecurityType.DRVT,
    KsProduct.ETF: SecurityType.ETF
}
PRODUCT_MY2KS = { v:k for k,v in PRODUCT_KS2MY.items() }

def symbol_ks2my(vt_symbol: str):
    if not vt_symbol:
        return ''
    symbol, ks_exchange = extract_vt_symbol(vt_symbol)
    return f'{MARKET_KS2MY.get(ks_exchange)}.{symbol}'

def symbol_my2ks(my_symbol: str):
    if not my_symbol:
        return ''
    symbol, ks_exchange = extract_my_symbol(my_symbol)
    return generate_vt_symbol(symbol, ks_exchange)

SUBTYPE_KS2MY = {
    KsSubscribeType.USER_ORDER: KsSubscribeType.USER_ORDER,
    KsSubscribeType.USER_TRADE: KsSubscribeType.USER_TRADE,
    KsSubscribeType.USER_POSITION: KsSubscribeType.USER_POSITION,
    KsSubscribeType.TRADE: SubType.TICKER,
    KsSubscribeType.BOOK: SubType.ORDER_BOOK,
    
    KsSubscribeType.K_MINUTE: SubType.K_1M,
    KsSubscribeType.K_HOUR: SubType.K_60M,
    KsSubscribeType.K_DAILY: SubType.K_DAY,
    KsSubscribeType.K_WEEK: SubType.K_WEEK,
    KsSubscribeType.K_MONTH: SubType.K_MON,

    KsSubscribeType.K_MINUTE5: SubType.K_5M,
    KsSubscribeType.K_MINUTE15: SubType.K_15M,
    KsSubscribeType.K_MINUTE30: SubType.K_30M
}

ADJUSTMENT_KS2MY = {
    KsAdjustment.BACKWARD_ADJUSTMENT: AuType.QFQ,
    KsAdjustment.FORWARD_ADJUSTMENT: AuType.HFQ,
    KsAdjustment.NONE: AuType.NONE
}

INTERVAL_KS2MY = {
    KsInterval.MINUTE: KLType.K_1M,
    KsInterval.HOUR: KLType.K_60M,
    KsInterval.DAILY: KLType.K_DAY,
    KsInterval.WEEK: KLType.K_WEEK,
    KsInterval.MONTH: KLType.K_MON,

    KsInterval.MINUTE5: KLType.K_5M,
    KsInterval.MINUTE15: KLType.K_15M,
    KsInterval.MINUTE30: KLType.K_30M
}

INTERVAL_MY2KS = {v:k for k,v in INTERVAL_KS2MY.items()}

EXCHANGE2TZ = {
    KsExchange.SEHK: CHINA_TZ,
    KsExchange.SMART: US_EASTERN_TZ
}

def extract_my_symbol(my_symbol: str):
    items = my_symbol.split('.')
    return '.'.join(items[1:]), MARKET_MY2KS.get(items[0])

def get_tz(vt_symbol: str):
    symbol, exchange = extract_vt_symbol(vt_symbol)
    return EXCHANGE2TZ.get(exchange)

class TimeoutException(Exception):
    def __init__(self, message: str, timeout: int):
        super().__init__(message)
        self.timeout = timeout


class KsFutuMarketApi(BaseMarketApi):
    gateway_name: str = 'KS_FUTU'

    def __init__(self, setting: dict):
        security_firm = setting.get('security_firm')
        self.port = setting.get('port', 11111)
        self.timeout = setting.get('timeout', 60)
        gateway_name = setting.get('gateway_name', self.gateway_name)
        dd_secret = setting.get('dd_secret')
        dd_token = setting.get('dd_token')
        super().__init__(gateway_name=gateway_name, dd_secret=dd_secret, dd_token=dd_token)

        self.init_handlers(security_firm)

        self.plugins = {}
        plugins_settings = setting.get('plugins', {})
        for name, plugin_setting in plugins_settings.items():
            plugin_class_name = plugin_setting['name'].title().replace('_', '')
            self.plugins[name] = globals()[plugin_class_name](plugin_setting['setting'])


    # 初始化行回调和订单回调
    def init_handlers(self, security_firm):
        trade = self

        connect_start = datetime.now()
        self.quote_ctx = quote_ctx = OpenQuoteContext(host='127.0.0.1', port=self.port, is_async_connect=True)
        while True:
            if self.quote_ctx._status == ContextStatus.READY:
                self.log('Futu OpenD连接成功')
                break
            if (datetime.now() - connect_start).seconds > self.timeout:
                self.quote_ctx.close()
                raise TimeoutException('OpenQuoteContext连接超时，请检查Futu OpenD是否打开', self.timeout)
            sleep(0.1)
        # self.trd_ctx = trd_ctx = OpenSecTradeContext(host='127.0.0.1', port=11111, filter_trdmarket=TrdMarket.NONE, security_firm=security_firm)  # 创建交易对象

        # 盘口 callback
        class OrderBookHandler(OrderBookHandlerBase):
            def on_recv_rsp(self, rsp_pb):
                ret_code, data = super(OrderBookHandler,self).on_recv_rsp(rsp_pb)
                if ret_code != RET_OK:
                    trade.log({'msg': data}, level=ERROR, name='on_order_book')
                    return ret_code, data
                
                book: MyBookData = trade.book_my2ks(data)
                # symbol, exchange = extract_my_symbol(data['code'])
                # book: MyBookData = MyBookData(
                #     symbol=symbol,
                #     exchange=exchange,
                #     datetime=datetimes.now(),
                #     name=symbol,
                #     gateway_name=trade.gateway_name
                # )

                # for index, bid_item in enumerate(data['Bid']):
                #     i = index + 1
                    
                #     bid_price = Decimal(str(bid_item[0]))
                #     bid_volume = Decimal(str(bid_item[1]))
                    
                #     setattr(book, f'bid_price_{i}', bid_price)
                #     setattr(book, f'bid_volume_{i}', bid_volume)
                   
                #     if data['Ask']:
                #         ask_item = data['Ask'][index]
                #         ask_price = Decimal(str(ask_item[0]))
                #         ask_volume = Decimal(str(ask_item[1]))
                #         setattr(book, f'ask_price_{i}', ask_price)
                #         setattr(book, f'ask_volume_{i}', ask_volume)

                trade.on_book(book)
                return RET_OK, data
        handler = OrderBookHandler()
        quote_ctx.set_handler(handler)
        
        # 分笔 callback
        class TickerHandler(TickerHandlerBase):
            def on_recv_rsp(self, rsp_pb):
                ret_code, data_df = super(TickerHandler,self).on_recv_rsp(rsp_pb)
                if ret_code != RET_OK:
                    trade.log({'msg': data_df}, level=ERROR, name='on_tick')
                    return RET_ERROR, data_df
                
                if not len(data_df):
                    return ret_code, data_df
                
                data = data_df.to_dict('records')[0]
                # trade.log(data, name='on_ticker')
                
                symbol, exchange = extract_my_symbol(data['code'])
                tz = CHINA_TZ if exchange == KsExchange.SEHK else US_EASTERN_TZ
                dt: datetime = tz.localize(parse(data['time'])).astimezone(CHINA_TZ)
                tick: MyRawTickData = MyRawTickData(
                    symbol=symbol,
                    exchange=exchange,
                    datetime=dt,
                    name=symbol,
                    volume=Decimal(str(data['volume'])),
                    turnover=Decimal(str(data['turnover'])),
                    last_price=Decimal(str(data['price'])),
                    gateway_name=trade.gateway_name
                )
                tick.last_volume = tick.volume
                trade.on_tick(tick)
                return RET_OK, data_df
        handler = TickerHandler()
        quote_ctx.set_handler(handler)

        class CurKlineHandler(CurKlineHandlerBase):
            def on_recv_rsp(self, rsp_pb):
                ret_code, data_df = super(CurKlineHandler,self).on_recv_rsp(rsp_pb)
                if ret_code != RET_OK:
                    trade.log({'msg': data_df}, level=ERROR, name='on_bar_rt')
                    return RET_ERROR, data_df
                
                data = data_df.to_dict('records')[0]

                symbol, exchange = extract_my_symbol(data['code'])
                vt_symbol = generate_vt_symbol(symbol, exchange)
                tz = get_tz(vt_symbol)
                data_df['datetime'] = pd.to_datetime(data_df.time_key).dt.tz_localize(tz)
                
                
                bar: BarData = BarData(
                    symbol=symbol,
                    exchange=exchange,
                    datetime=data_df['datetime'].iloc[0],
                    
                    interval = INTERVAL_MY2KS.get(data_df['k_type'].iloc[0]),
                    volume = Decimal(str(data_df['volume'].iloc[0])),
                    turnover = Decimal(str(data_df['turnover'].iloc[0])),
                    open = Decimal(str(data_df['open'].iloc[0])),
                    high = Decimal(str(data_df['high'].iloc[0])),
                    low = Decimal(str(data_df['low'].iloc[0])),
                    close = Decimal(str(data_df['close'].iloc[0])),

                    gateway_name=trade.gateway_name
                )
                trade.on_bar_rt(bar)
                return RET_OK, data_df
        handler = CurKlineHandler()
        quote_ctx.set_handler(handler)

        

    # 订阅行情
    def subscribe(self, vt_symbols, vt_subtype_list, extended_time=True) -> tuple[KsRetCode, Optional[ErrorData]]:
        if not vt_symbols:
            return KS_RET_OK, None
        
        if isinstance(vt_symbols, str):
            vt_symbols = [vt_symbols]

        my_symbols = [symbol_ks2my(x) for x in vt_symbols]
        my_subtype_list = [SUBTYPE_KS2MY.get(x) for x in vt_subtype_list]

        trade = self
        if KsSubscribeType.USER_ORDER in my_subtype_list:
            my_subtype_list.remove(KsSubscribeType.USER_ORDER)     

        if KsSubscribeType.USER_TRADE in my_subtype_list:
            my_subtype_list.remove(KsSubscribeType.USER_TRADE)

        # futu没有持仓回调
        if KsSubscribeType.USER_POSITION in my_subtype_list:
            my_subtype_list.remove(KsSubscribeType.USER_POSITION)

        # 剩下的是订阅行情
        if my_subtype_list:
            ret, data = self.quote_ctx.subscribe(my_symbols, my_subtype_list, extended_time=extended_time)   # 订阅 K 线数据类型，FutuOpenD 开始持续收到服务器的推送
            if ret == RET_ERROR:
                error = self.get_error(vt_symbols, vt_subtype_list, extended_time, code=KsErrorCode.SUBSCRIPTION_ERROR, msg=data)
                self.send_dd(error.msg, f'订阅行情错误')
                return KS_RET_ERROR, error
            return KS_RET_OK, data

        return KS_RET_OK, None
    
    def unsubscribe(self, vt_symbols, vt_subtype_list) -> tuple[KsRetCode, Optional[ErrorData]]:
        if not vt_symbols:
            return KS_RET_OK, None
        
        if isinstance(vt_symbols, str):
            vt_symbols = [vt_symbols]

        my_symbols = [symbol_ks2my(x) for x in vt_symbols]
        my_subtype_list = [SUBTYPE_KS2MY.get(x) for x in vt_subtype_list]


        # 剩下的是订阅行情
        if my_subtype_list:
            ret, data = self.quote_ctx.unsubscribe(my_symbols, my_subtype_list)   # 订阅 K 线数据类型，FutuOpenD 开始持续收到服务器的推送
            if ret == RET_ERROR:
                error = self.get_error(vt_symbols, vt_subtype_list, code=KsErrorCode.SUBSCRIPTION_ERROR, msg=data)
                self.send_dd(error.msg, f'反订阅行情错误')
                return KS_RET_ERROR, error
            return KS_RET_OK, data

        return KS_RET_OK, None

              
    def get_error(self, *args, **kvargs):
        method = sys._getframe(1).f_code.co_name
        error = ErrorData(
            code=kvargs.get('code'),
            msg=kvargs.get('msg'),
            method=method,
            args=args,
            kvargs=kvargs,
            traceback=traceback.format_exc(),
            gateway_name=self.gateway_name
        )
        self.log(error, tag=f'api_error.{method}', level=ERROR)
        return error

    # 获取静态信息
    def query_contract(self, vt_symbol: str) -> tuple[KsRetCode, ContractData]:
        symbol, exchange = extract_vt_symbol(vt_symbol)
        my_symbol = symbol_ks2my(vt_symbol)
        ret, data = self.quote_ctx.get_stock_basicinfo(MARKET_KS2MY.get(exchange), code_list=[my_symbol])
        if ret == RET_ERROR:
            error = self.get_error(vt_symbol, data, msg=data)
            return KS_RET_ERROR, error
        
        contract_data = data.iloc[0]
        product: KsProduct = PRODUCT_MY2KS.get(contract_data.stock_type)
        lot_size: Decimal = Decimal(str(contract_data.lot_size))
        is_stock: bool = is_stock_like(product)
        size: int = Decimal('1') if is_stock else lot_size
        min_volume: int = lot_size if is_stock else Decimal('1')
        contract = ContractData(
            symbol=symbol,
            exchange=exchange,
            product=product,
            size=size,
            min_volume=min_volume,
            # pricetick=Decimal('0.01'), # todo! 低价股是0.001这里以后要处理
            name=contract_data.get('name'),
            gateway_name=self.gateway_name
        )
        return KS_RET_OK, contract
    
    # 获取静态信息 # todo! ks_trader_wrapper中使用到df=False要修正那边
    def query_contracts(
            self,
            vt_symbols: Optional[List[str]] = None,
            exchanges: Optional[list[KsExchange]] = None,
            products: Optional[List[KsProduct]] = None,
            df: bool = True
        ) -> tuple[KsRetCode, Union[list[ContractData], DataFrame]]:
        #  特殊处理CN，把深市和沪市合并
        if exchanges and KsExchange.CNSE in exchanges:
            exchanges.remove(KsExchange.CNSE)
            exchanges += [KsExchange.SSE, KsExchange.SZSE]
        

        if vt_symbols:
            my_symbols = [symbol_ks2my(x) for x in vt_symbols]
            ret, data = self.quote_ctx.get_stock_basicinfo('US', code_list=my_symbols) # futu接口如果指定code_list，会忽略exchange
            if ret == RET_ERROR:
                error = self.get_error(vt_symbols, data, msg=data)
                return KS_RET_ERROR, error
        else:
            data = pd.DataFrame()
            my_products = [PRODUCT_KS2MY.get(x) for x in products]
            my_exchanges = [MARKET_KS2MY.get(x) for x in exchanges]
            for product in my_products:
                for exchange in my_exchanges:
                    ret, data1 = self.quote_ctx.get_stock_basicinfo(market=exchange, stock_type=product)
                    if ret == RET_ERROR:
                        error = self.get_error(vt_symbols, exchanges, products, df, msg=data1)
                        return KS_RET_ERROR, error
                    data = pd.concat([data, data1], ignore_index=True)

        if len(data):
            # 过滤掉退市标的
            data = data[data.delisting==False]
        
        if df:
            data['vt_symbol'] = [symbol_my2ks(x) for x in data['code']]
            data['product'] = [PRODUCT_MY2KS.get(x).name for x in data['stock_type']]
            data['size'] = '1'
            data['min_volume'] = data['lot_size'].astype(str)
            data['gateway'] = self.gateway_name
            data['sub_exchange'] = data['exchange_type']
            return KS_RET_OK, data[['vt_symbol', 'name', 'product', 'size', 'min_volume', 'sub_exchange', 'gateway']]
        
        contracts: list[ContractData] = []
        for index, contract_data in data.iterrows():
            symbol, exchange = extract_my_symbol(contract_data.code)
            contract = ContractData(
                symbol=symbol,
                exchange=exchange,
                sub_exchange=contract_data.exchange_type,
                product=KsProduct.EQUITY,
                size=Decimal('1'),
                min_volume=Decimal(str(contract_data.lot_size)),
                # pricetick=Decimal('0.01'), # todo! 低价股是0.001这里以后要处理
                name=contract_data.get('name'),
                gateway_name=self.gateway_name
            )
            contract.exchange_type = contract_data.exchange_type
            contracts.append(contract)
        return KS_RET_OK, contracts
    
    def query_book(self, vt_symbol: str) -> tuple[KsRetCode,  MyBookData]:
        my_symbol = symbol_ks2my(vt_symbol)
        ret_sub, sub_data = self.quote_ctx.subscribe([my_symbol], [SubType.ORDER_BOOK], subscribe_push=False)
        # 先订阅买卖摆盘类型。订阅成功后 OpenD 将持续收到服务器的推送，False 代表暂时不需要推送给脚本
        ret_code = RET_ERROR
        ret_data = None
        if ret_sub == RET_OK:  # 订阅成功
            ret, data = self.quote_ctx.get_order_book(my_symbol, num=5)  # 获取一次 3 档实时摆盘数据
            if ret == RET_OK:
                ret_code = KS_RET_OK
                ret_data = self.book_my2ks(data)
            else:
                ret_data = ret_data
        else:
            ret_data = sub_data
        return ret_code, ret_data
    
    def book_my2ks(self, data) -> MyBookData:
        symbol, exchange = extract_my_symbol(data['code'])
        book: MyBookData = MyBookData(
            symbol=symbol,
            exchange=exchange,
            datetime=datetimes.now(),
            name=symbol,
            gateway_name=self.gateway_name
        )

        for index, bid_item in enumerate(data['Bid']):
            i = index + 1
            
            bid_price = Decimal(str(bid_item[0]))
            bid_volume = Decimal(str(bid_item[1]))
            
            setattr(book, f'bid_price_{i}', bid_price)
            setattr(book, f'bid_volume_{i}', bid_volume)
            
            if data['Ask'] and index < len(data['Ask']):
                ask_item = data['Ask'][index]
                ask_price = Decimal(str(ask_item[0]))
                ask_volume = Decimal(str(ask_item[1]))
                setattr(book, f'ask_price_{i}', ask_price)
                setattr(book, f'ask_volume_{i}', ask_volume)
        return book
    
    # todo! ks_trader_rapper没有适配好
    def query_quotes(self, vt_symbols: list[str], df: bool = True) -> Union[KsRetCode, list[QuoteData]]:
        if not vt_symbols:
            return KsRetCode, []
        
        my_symbols = [symbol_ks2my(x) for x in vt_symbols]
        ret, data = self.quote_ctx.get_market_snapshot(my_symbols)

        # ETF没有市值，只有净资产
        data['circular_market_val'] = np.where(data['trust_valid'], data['trust_aum'], data['circular_market_val'])
        data['total_market_val'] = data['circular_market_val']

        if ret == RET_OK:
            if df:
                tz = get_tz(vt_symbols[0])
                data['vt_symbol'] = [symbol_my2ks(x) for x in data['code']]
                data['datetime'] = pd.to_datetime(data['update_time']).dt.tz_localize(tz)
                data['volume'] = data['volume'].astype(str)
                data['turnover'] = data['turnover'].astype(str)
                data['turnover_rate'] = (data['turnover_rate']/100).astype(str)
                data['last_price'] = data['last_price'].astype(str)
                data['open_price'] = data['open_price'].astype(str)
                data['high_price'] = data['high_price'].astype(str)
                data['low_price'] = data['low_price'].astype(str)
                data['pre_close'] = data['prev_close_price'].astype(str)
                data['circular_shares'] = data['outstanding_shares'].astype(str)
                data['total_shares'] = data['issued_shares'].astype(str)
                data['circular_market_cap'] = data['circular_market_val'].astype(str)
                data['total_market_cap'] = data['total_market_val'].astype(str)
                data['pre_close'] = data['prev_close_price'].astype(str)

                # fundalmental
                data['net_asset'] = data['net_asset'].astype(str)
                data['net_profit'] = data['net_profit'].astype(str)
                data['net_asset_per_share'] = data['net_asset_per_share'].astype(str)
                data['dividend_ratio_ttm'] = (data['dividend_ratio_ttm']/100).astype(str)
                data['pb_ratio'] = (data['pb_ratio']/100).astype(str)
                data['pe_ratio'] = (data['pe_ratio']/100).astype(str)
            
                return KS_RET_OK, data[[
                      'vt_symbol', 'datetime', 'volume', 'turnover', 'turnover_rate', 'last_price',
                      'open_price', 'high_price', 'low_price', 'pre_close', 'circular_shares', 'total_shares', 
                      'circular_market_cap', 'total_market_cap', 'net_asset', 'net_profit', 'net_asset_per_share', 
                      'dividend_ratio_ttm', 'pb_ratio', 'pe_ratio'
                    ]]
            else:
                quotes = []
                for index, quote in data.iterrows():
                    symbol, exchange = extract_my_symbol(quote.code)
                    tz = CHINA_TZ if exchange == KsExchange.SEHK else US_EASTERN_TZ
                    dt = tz.localize(parse(f'{quote.data_date} {quote.data_time}')).astimezone(CHINA_TZ)
                    quotes.append(QuoteData(
                        gateway_name=self.gateway_name,
                        symbol=symbol,
                        exchange=exchange,
                        datetime=dt,
                        volume=Decimal(quote.volume),
                        turnover=Decimal(quote.turnover),
                        turnover_rate=Decimal(quote.turnover_rate/100),
                        last_price=quote.last_price,
                        open_price=quote.open_price,
                        high_price=quote.high_price,
                        low_price=quote.low_price,
                        pre_close=quote.prev_close_price,
                        circular_shares=quote.outstanding_shares,
                        total_shares=quote.issued_shares,
                        circular_market_cap=quote.circular_market_val,
                        total_market_cap=quote.total_market_val,
                        localtime=datetimes.now()
                    ))
                return KS_RET_OK, quotes
        else:
            return KS_RET_ERROR, self.get_error(vt_symbols=vt_symbols, msg=data)
        
    def query_ticks(self, vt_symbol: str, length: int = 1) -> Union[KsRetCode, list[MyRawTickData]]:
        if not vt_symbol:
            return KsRetCode, []
        
        my_symbol = symbol_ks2my(vt_symbol)
        ret_sub, data_sub = self.quote_ctx.subscribe([my_symbol], [SubType.TICKER], subscribe_push=False, extended_time=True)
        # 先订阅 K 线类型。订阅成功后 OpenD 将持续收到服务器的推送，False 代表暂时不需要推送给脚本
        if ret_sub == RET_OK:  # 订阅成功
            ret_quote, data_quote = self.quote_ctx.get_rt_ticker(my_symbol, length)  # 获取订阅股票报价的实时数据
            quotes: list[QuoteData] = []
            if ret_quote == RET_OK:
                for index, quote in data_quote.iterrows():
                    symbol, exchange = extract_my_symbol(quote.code)
                    tz = CHINA_TZ if exchange == KsExchange.SEHK else US_EASTERN_TZ
                    dt = tz.localize(parse(f'{quote.time}')).astimezone(CHINA_TZ)
                    quotes.append(MyRawTickData(
                        gateway_name=self.gateway_name,
                        symbol=symbol,
                        exchange=exchange,
                        datetime=dt,
                        volume=Decimal(str(quote.volume)),
                        last_price=Decimal(str(quote.price))
                    ))
                return KS_RET_OK, quotes
            else:
                return KS_RET_ERROR, self.get_error(vt_symbol=vt_symbol, msg=data_quote)
        else:
            return KS_RET_ERROR, self.get_error(vt_symbol=vt_symbol, msg=data_sub)
    
    # https://openapi.futunn.com/futu-api-doc/quote/get-kl.html
    # 由于实际上的历史k线获取收到额度限制，这里用实时K线来代替历史K线，实时K线需要先订阅行情，否则会出错

    # todo 这里默认先不管是不是df都返回df，后续在处理非df 
    def query_history_n(
            self,
            vt_symbol: str,
            n: int,
            interval: KsInterval,
            adjustment: KsAdjustment,
            df: bool = True
        ) -> tuple[KsRetCode, Union[list[BarData], DataFrame]]:
        """
        Query bar history data.
        """
        my_symbol = symbol_ks2my(vt_symbol)
        # 结果之后实时K线会返回当天最新的，要去除掉
        ret_k, data_k = self.quote_ctx.get_cur_kline(my_symbol, n+1, ktype=INTERVAL_KS2MY.get(interval), autype=ADJUSTMENT_KS2MY.get(adjustment))
        if not ret_k == RET_OK:
            return KS_RET_ERROR, self.get_error(vt_symbol=vt_symbol, n=n, interval=interval, adjustment=adjustment, msg=data_k)
        tz = get_tz(vt_symbol)
        
        data_k['datetime'] = pd.to_datetime(data_k.time_key).dt.tz_localize(tz)
        
        # 如果是盘中时间就需要-1，否则不需要
        symbol, exchange = extract_vt_symbol(vt_symbol)
        start_time, end_time = get_trading_time(exchange, extended_time=True)
        now_dt = datetimes.now(tz=tz)
        if start_time <= now_dt <= end_time:
            data_k = data_k[:-1]
        
        data_k['vt_symbol'] = vt_symbol
        
        data_k['interval'] = interval.value
        data_k['volume'] = data_k.volume
        data_k['turnover'] = data_k.turnover
        data_k['open'] = data_k.open
        data_k['high'] = data_k.high
        data_k['low'] = data_k.low
        data_k['close'] = data_k.close
        ret_df = data_k[['vt_symbol', 'datetime', 'interval', 'volume', 'turnover', 'open', 'high', 'low', 'close']]
        return KS_RET_OK, ret_df

    # 关闭上下文连接
    def close(self):
        self.quote_ctx.close()


        


from ks_utility.zmqs import ZmqPublisher, ZmqSubscriber
from ks_trade_api.constant import SubscribeType, Indicator, Timing, RET_OK, Exchange, Interval
from ks_trade_api.object import IndicatorData
from dateutil.parser import parse
import json
import traceback

class KsMarketApi(ZmqPublisher, ZmqSubscriber):
    gateway_name: str = 'ks_market_api'
    
    def __init__(self, setting: dict = {}):
        self.gateway_name = setting.get('gateway_name', self.gateway_name)
        
        pub_address: str = setting.get('zmq', {}).get('pub_address')
        sub_address: str = setting.get('zmq', {}).get('sub_address')
        if not sub_address:
            from config import GATEWAY_CONFIG
            sub_address = GATEWAY_CONFIG['ks_market_api']['setting']['zmq']['sub_address']
        if not pub_address:
            pub_address = GATEWAY_CONFIG['ks_market_api']['setting']['zmq']['pub_address']
        ZmqPublisher.__init__(self, pub_address)
        ZmqSubscriber.__init__(self, sub_address)

    def subscribe(
            self,
            vt_symbols: list[str] = [], 
            types: list[SubscribeType] = [], 
            indicators: list[Indicator] = [],
            data_time_types: list[Timing] = []
        ):
        types = [x.value for x in types]
        indicators = [x.value for x in indicators]
        data_time_types = [x.value for x in data_time_types]
        self.send('subscribe', {'vt_symbols': vt_symbols, 'types': types, 'indicators': indicators, 'data_time_types': data_time_types})
        return RET_OK, None

    def on_indicator(self, indicator):
        pass
    
    def on_close(self, data: dict = {}):
        pass

    def on_message(self, topic: str, msg: str):
        msg_data = json.loads(msg)
        if topic == 'on_close':
            self.on_close(msg_data)
            return
        
        try:
            indicator = IndicatorData(
                gateway_name=msg_data['gateway_name'],
                symbol=msg_data['symbol'],
                exchange=Exchange(msg_data['exchange']),
                datetime=parse(msg_data['datetime']),
                indicator=Indicator(msg_data['indicator']),
                interval=Interval(msg_data['interval']),
                timing=Timing(msg_data['timing']),
                data=msg_data['data']  
            )
            getattr(self, topic)(indicator)
        except:
            traceback.print_exc()
            
    def close(self):
        self.send('close', {})
        ZmqPublisher.stop(self)
        ZmqSubscriber.stop(self)

if __name__ == '__main__':
    ks_market_api = KsMarketApi()



import multiprocessing
import time
import random
import os


class BarData:
    """ 保存股票数据 """
    def __init__(self, price, volume, symbol, period):
        self.price = price
        self.volume = volume
        self.symbol = symbol
        self.period = period  # ✅ 每条数据带上周期

    def __repr__(self):
        return f"BarData(price={self.price}, volume={self.volume}, symbol={self.symbol}, period={self.period})"


def process_b(bar_data, shared_dict):
    """ 计算不同周期的 price 均值 """
    pid = os.getpid()  # 获取当前进程 ID
    symbol = bar_data.symbol
    period = bar_data.period  # ✅ 取出当前数据的周期

    # ✅ 确保 symbol 在共享数据中
    if symbol not in shared_dict:
        print(f"[进程 {pid}] {symbol} 不在 shared_dict 中，数据丢失！")
        return

    # ✅ 确保 symbol 下有该周期的价格列表
    if period not in shared_dict[symbol]:
        shared_dict[symbol][period] = multiprocessing.Manager().list()

    price_list = shared_dict[symbol][period]
    price_list.append(bar_data.price)

    if len(price_list) > period:
        price_list.pop(0)  # 只保留最近 `period` 个价格

    avg_price = sum(price_list) / len(price_list)

    print(f"[进程 {pid}] 股票: {symbol} | 周期: {period} | 最近 {period} 个 price: {list(price_list)} | 均值: {avg_price:.2f}")


if __name__ == "__main__":
    stock_symbols = ["AAPL", "TSLA", "AMZN", "MSFT"]
    periods = [3, 5, 10]  # ✅ 支持多个周期
    pool_size = 4  

    manager = multiprocessing.Manager()
    shared_dict = manager.dict()  # 使用共享字典存储 {symbol: {周期1: [最近N个price], 周期2: [...]}}
    
    # ✅ 初始化 shared_dict
    for symbol in stock_symbols:
        shared_dict[symbol] = manager.dict()  
        for period in periods:
            shared_dict[symbol][period] = manager.list()  # 每个周期都有自己的价格列表

    pool = multiprocessing.Pool(processes=pool_size)

    try:
        while True:
            # 生成随机股票数据
            symbol = random.choice(stock_symbols)
            price = random.randint(100, 500)
            volume = random.randint(10, 100)
            period = random.choice(periods)  # ✅ 随机选择一个周期

            bar_data = BarData(price, volume, symbol, period)
            pool.apply_async(process_b, (bar_data, shared_dict))  # 提交任务到进程池

            time.sleep(0.5)  # 模拟 on_bar 调用频率

    except KeyboardInterrupt:
        print("主进程退出，关闭进程池...")
        pool.close()
        pool.join()



from ks_trade_api.utility import load_json, TEMP_NAME
from typing import Dict, Any
from logging import INFO
from pathlib import Path
import os
import sys

SETTINGS: Dict[str, Any] = {
    "ks_market_api": {
        "name": "ks_market_api",
        "setting": {
            "security_firm": "FUTUSG",
            "port": 11111,
            "zmq": {
                "pub_address": "tcp://127.0.0.1:2000",
                "sub_address": "tcp://127.0.0.1:2001"
            }
        }
    }
}


# Load global setting from json file.
SETTING_FILENAME: str = "ks_setting.json"
SETTINGS.update(load_json(SETTING_FILENAME))
